'use strict';
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var print_1 = __importDefault(require("./lib/print"));
var suffixCheck_1 = __importDefault(require("./helpers/suffixCheck"));
var optionalReplace_1 = require("./helpers/optionalReplace");
var customNameReplace_1 = __importDefault(require("./helpers/customNameReplace"));
var typeOf_1 = __importDefault(require("./helpers/typeOf"));
var list_1 = __importDefault(require("./helpers/list"));
var msgPOJO_1 = __importDefault(require("./core/msgPOJO"));
var primitives_1 = __importDefault(require("./core/primitives"));
var NotTypeError_1 = __importDefault(require("./core/NotTypeError"));
var You = {
    timestamp: false,
    messageInPOJO: false,
    _primitives: primitives_1.default,
    _lodged: []
    // Opinions
    ,
    opinionatedOnNaN: true,
    opinionatedOnArray: true,
    opinionatedOnNull: true,
    _isOpinionated: true
    // isOpinionated getters and setters
    // A "gang" value that will flip all opinions on/off
    ,
    get isOpinionated() { return this._isOpinionated; },
    set isOpinionated(value) {
        this.opinionatedOnNaN = value;
        this.opinionatedOnArray = value;
        this.opinionatedOnNull = value;
        this._isOpinionated = value;
    }
    // Define whether will throw errors (true by default)
    ,
    willThrowError: true,
    get throw() { return this.willThrowError; },
    set throw(value) {
        this.willThrowError = value;
    }
    // verbosity
    ,
    get verbose() { return this.timestamp && this.messageInPOJO; },
    set verbose(value) {
        this.timestamp = this.messageInPOJO = value;
    }
    // METHODS
    // You.#are is the core method for validation
    ,
    are: function (expect, got, name, note) {
        return !this.areNot(expect, got);
    }
    // _are is validation help for internal use. Has a try/catch wrapper to prevent throwing errors.
    ,
    _are: function (expect, got, name, note) {
        try {
            var fail = this.areNot(expect, got, name, note);
            if (typeof fail === 'string')
                return false;
            return true;
        }
        catch (error) {
            return false;
        }
    }
    // You.#areNot is the core method for negative validation
    ,
    areNot: function (expect, got, name, note) {
        expect = this.prepareExpect(expect);
        var gotType = typeOf_1.default(got, this);
        if (this.validate(expect, got, gotType))
            return false;
        var message = this.msg(expect, got, gotType, name, note);
        if (this.willThrowError) {
            var error;
            if (typeof message === 'object') {
                error = new NotTypeError_1.default(message.message);
                error.trace = message;
            }
            else {
                error = new NotTypeError_1.default(message);
            }
            throw error;
        }
        return message;
    }
    // create is the default way to "instantiate"
    ,
    create: function (options) {
        var you = Object.create(this);
        you._lodged = [];
        if (options)
            this._applyOptions(you, options);
        return you;
    },
    createNot: function (options) {
        var you = Object.create(this);
        you._lodged = [];
        if (options)
            this._applyOptions(you, options);
        return you.not.bind(you);
    }
    // createIs is the default way to make a simplified true/false method.
    // TODO why you.not
    ,
    createIs: function (options) {
        var you = Object.create(this);
        this._applyOptions(you, options);
        return you.are.bind(you);
    }
    // checkObject allows checking of typing with an object literal `{}`.
    ,
    checkObject: function (name, expectObject, gotObject, callback) {
        // necessary run-time type-checking.
        this.areNot('string', name);
        this.areNot('object', expectObject);
        this.areNot('object', gotObject);
        this.areNot(['function', 'object', 'optional'], callback);
        // set up another object to not pollute the current one.
        var not = Object.create(this);
        if (typeof callback === 'function') {
            not.walkObject(name, expectObject, gotObject);
            return not.resolve(callback, null); // null to specify no payload
        }
        if (typeof callback === 'object') {
            var returnedPayload = null;
            // walk payload
            if (callback.returnPayload === true) {
                returnedPayload = not.walkObject(name, expectObject, gotObject, true, callback.exact);
                if (returnedPayload === '$$empty$$')
                    returnedPayload = null;
            }
            else {
                not.walkObject(name, expectObject, gotObject, null, callback.exact);
            }
            // set callback
            if (typeof callback.callback === 'function') {
                callback = callback.callback;
            }
            else {
                if (this.throw) {
                    // if callback is not defined
                    // set a default callback to throw errors
                    callback = function (errors, payload) {
                        if (errors) {
                            var toThrow = new NotTypeError_1.default('Wrong types provided. See `trace`.');
                            if (typeof errors !== 'boolean')
                                toThrow.trace = errors;
                            throw toThrow;
                        }
                        return payload;
                    };
                }
                else {
                    callback = function (errors, payload) {
                        if (errors)
                            return errors;
                        return payload;
                    };
                }
            }
            return not.resolve(callback, returnedPayload);
        }
        not.walkObject(name, expectObject, gotObject);
        return not.resolve();
    }
    // customExpectHdlr handles custom types.
    ,
    customExpectHdlr: function (r, expect) {
        if (this._primitives.indexOf(expect) === -1) {
            if (this["$$custom_" + expect] !== undefined) {
                r.push("$$custom_" + expect);
                return r;
            }
            throw new NotTypeError_1.default("Internal error: `" + expect + "` is not a valid type to check for.");
        }
        r.push(expect);
        return r;
    }
    // defineType allows users to define their own types/validation.
    ,
    defineType: function (payload) {
        var _this = this;
        var sanitised = You.checkObject('defineType', {
            primitive: ['string', 'array'],
            type: 'string',
            pass: ['function', 'optional']
        }, payload, { returnPayload: true });
        if (Array.isArray(sanitised)) {
            sanitised.forEach(function (el) {
                console.error(el);
            });
            throw new NotTypeError_1.default('Wrong inputs for #defineType.');
        }
        if (typeof sanitised.primitive === 'string')
            sanitised.primitive = [sanitised.primitive];
        sanitised.primitive.forEach(function (p) {
            if (_this._primitives.indexOf(p) === -1)
                throw new NotTypeError_1.default("Internal error: `" + p + "` is not a valid primitive.");
        });
        var key = "$$custom_" + sanitised.type;
        this[key] = { primitive: sanitised.primitive };
        if (sanitised.pass)
            this[key]['pass'] = sanitised.pass;
    }
    // Messaging generator
    ,
    msg: function (expect, got, gotType, name, note) {
        var gotTypeListed = list_1.default(gotType);
        var msg = 'Wrong Type';
        msg += name ? " (" + name + ")" : '';
        msg += ": Expecting type " + list_1.default(expect) + " but got " + gotTypeListed;
        // no need to elaborate for null, undefined and nan
        msg += (['`null`', '`undefined`', '`nan`'].indexOf(gotTypeListed) > -1) ? '.' : " with value of `" + print_1.default(got) + "`.";
        msg += note ? " Note: " + note + "." : '';
        if (this.timestamp)
            msg += " (TS: " + new Date().getTime() + ")";
        if (this.messageInPOJO)
            return msgPOJO_1.default(msg, expect, got, gotType, name, note);
        return msg;
    },
    msgProp: function (expect, got, name, note) {
        var msg = name ? " (" + name + ")" : '';
        msg += "Property";
        if (expect) {
            msg += " `" + expect + "` is missing.";
        }
        else {
            msg += " `" + got + "` exists but is not expected.";
        }
        msg += note ? " Note: " + note + "." : '';
        if (this.timestamp)
            msg += " (TS: " + new Date().getTime() + ")";
        if (!this.willThrowError && this.messageInPOJO)
            return msgPOJO_1.default(msg, expect, got, undefined, name, note);
        return msg;
    }
    // #validate searches for the function to pass
    ,
    validate: function (expect, got, gotType) {
        if (typeof gotType == 'string')
            gotType = [gotType];
        var validated = false;
        for (var i = 0; i < expect.length; i++) {
            var el = expect[i];
            if (el.indexOf('$$') > -1) {
                // the customs must pass or fail as a whole, not in part.
                var passing = this._are(this[el].primitive, got, customNameReplace_1.default(el));
                if (!passing) {
                    continue; // if it doesn't pass the primitives check, no need to check further
                }
                else if (typeof this[el].pass === 'function') {
                    if (this[el].pass(got)) {
                        validated = true; // if there is a pass function, must pass it
                        break;
                    }
                    else {
                        continue;
                    }
                }
                else {
                    // there is no pass function to run. reaching here means it has passed.
                    validated = true;
                    break;
                }
            }
            else if (gotType.indexOf(el) !== -1) {
                validated = true;
                break;
            }
        }
        return validated;
    }
    // #lodge is used for handling multiple errors, by storing them first in #_lodged[].
    ,
    lodge: function (expect, got, name, note) {
        if (!this.hasOwnProperty('_lodged'))
            this._lodged = [];
        var _lodged = this._lodged;
        var lodge = false;
        // don't let lodge throw error
        try {
            lodge = this.areNot(expect, got, name, note);
        }
        catch (error) {
            lodge = error.message;
        }
        if (lodge)
            _lodged.push(lodge);
        return lodge;
    }
    // #lodgeProp is used for handling multiple errors in missing or additional properties, by storing them first in #_lodged[].
    ,
    lodgeProp: function (expect, got, name, note) {
        if (!this.hasOwnProperty('_lodged'))
            this._lodged = [];
        var lodge = this.msgProp(expect, got, name, note);
        this._lodged.push(lodge);
        return lodge;
    }
    // prepareExpect is an abstraction that handles the `expect` argument passed in, allowing `expect` to be `string` or `array`.
    ,
    prepareExpect: function (expect) {
        var _this = this;
        if (typeof expect === 'string') {
            expect = [expect];
        }
        else if (!Array.isArray(expect)) {
            var msg = "Internal error: Say what you expect to check as a string or array of strings.";
            msg += " Found " + list_1.default(typeOf_1.default(expect, this), 'as') + ".";
            throw new NotTypeError_1.default(msg);
        }
        var reducer = function (r, expect) {
            if (typeof expect !== 'string')
                throw new NotTypeError_1.default("Internal error: Say what you expect to check as a string. Found " + list_1.default(typeOf_1.default(expect, _this), 'as') + ".");
            expect = expect.toLowerCase();
            return _this.customExpectHdlr(r, expect);
        };
        return expect.reduce(reducer, []);
    }
    // #resolve is used when checks are completed, and we are ready to give all the errors back.
    ,
    resolve: function (callback, returnedPayload) {
        var _a = this, _lodged = _a._lodged, willThrowError = _a.willThrowError;
        if (_lodged.length === 0) {
            return (typeof callback === 'function') ? callback(false, returnedPayload) : false;
        }
        var lodged = _lodged.slice();
        this._lodged = []; // we need to use `this` to clear out the array by side-effects.
        if (typeof callback === 'function')
            return callback(lodged, returnedPayload);
        if (willThrowError) {
            var error = new NotTypeError_1.default('Wrong types provided. See `trace`.');
            error.trace = lodged;
            throw error;
        }
        return lodged;
    }
    // walkObject recursively goes through the object literal passed in to conduct checking
    ,
    walkObject: function (name, expectObject, gotObject, returnPayload, exact) {
        var _this = this;
        if (returnPayload)
            var sanitisedPayload = {};
        var expectKeys = Object.keys(expectObject);
        // if we want to force the object to be exact
        // we will need to point out the additional properties on the payload
        // for missing properties, the code block below will take care of them
        if (exact) {
            var gotKeys = Object.keys(gotObject);
            var additionals = gotKeys.filter(function (gotKey) { return !expectKeys.includes(gotKey); });
            additionals.forEach(function (additional) {
                _this.lodgeProp(false, "" + (name ? name + '.' : '') + additional);
            });
        }
        for (var i = 0, keys = expectKeys; i < keys.length; i++) {
            var key = keys[i];
            var expect = expectObject[key];
            var optional = false;
            var optionalString = '__optional';
            if (suffixCheck_1.default(key, '?')) {
                optional = optionalReplace_1.optionalReplace('?');
            }
            else if (suffixCheck_1.default(key, optionalString)) {
                optional = optionalReplace_1.optionalReplace(optionalString);
            }
            var keyCopy = optional ? optional(key) : key;
            // if the key is missing and is compulsory
            if (!(keyCopy in gotObject) && !optional) {
                this.lodgeProp("" + (name ? name + '.' : '') + keyCopy);
                continue;
            }
            var got = gotObject[keyCopy];
            // if object, walk further in
            // using typeof and other stuff for speed
            if (typeof expect === 'object' && expect !== null && !Array.isArray(expect)) {
                if (typeof got === 'object' && got !== null && !Array.isArray(expect)) {
                    var chunk = this.walkObject("" + (name ? name + '.' : '') + keyCopy, expect, got, returnPayload);
                    if (chunk === '$$empty$$')
                        continue;
                    if (returnPayload) {
                        sanitisedPayload[keyCopy] = got;
                    }
                    continue;
                }
                else {
                    if (optional)
                        continue;
                }
                this.lodge('object', got, "" + (name ? name + '.' : '') + keyCopy);
                continue;
            }
            if (optional) {
                if (Array.isArray(expect)) {
                    expect.push('optional');
                }
                else {
                    expect = [expect, 'optional'];
                }
            }
            var fail = this.lodge(expect, got, "" + (name ? name + '.' : '') + keyCopy);
            if (returnPayload && !fail && got)
                sanitisedPayload[keyCopy] = got;
        }
        if (returnPayload)
            return (Object.keys(sanitisedPayload).length < 1) ? '$$empty$$' : sanitisedPayload;
    }
    // _applyOptions is a helper to set up the options of descendant objects
    ,
    _applyOptions: function (descendant, options) {
        var _this = this;
        if (this._are('object', options)) {
            // Deprecated: willThrowError 
            if (this._are('boolean', options.willThrowError)) {
                console.warn('NotJS: Use of `willThrowError` is deprecated. Use `throw` insteead.');
                descendant.willThrowError = options.willThrowError;
            }
            if (this._are('boolean', options.throw))
                descendant.throw = options.throw;
            if (this._are('boolean', options.verbose))
                descendant.verbose = options.verbose;
            if (this._are('boolean', options.timestamp))
                descendant.timestamp = options.timestamp;
            if (this._are('boolean', options.messageInPOJO))
                descendant.messageInPOJO = options.messageInPOJO;
            if (this._are('boolean', options.isOpinionated)) {
                descendant.isOpinionated = options.isOpinionated;
                return;
            }
            Object.keys(You).filter(function (key) { return key.indexOf('opinionatedOn') === 0; }).forEach(function (optionKey) {
                if (_this._are('boolean', options[optionKey]))
                    descendant[optionKey] = options[optionKey];
            });
        }
    }
    // Additional overloads for semantics
    ,
    not: function (expect, got, name, note) {
        return this.areNot(expect, got, name, note);
    },
    is: function (expect, got, name, note) {
        return this.are(expect, got, name, note);
    }
    // scrub is a shorthand of #checkObject
    ,
    scrub: function (name, expectObject, gotObject, exact) {
        return this.checkObject(name, expectObject, gotObject, { returnPayload: true, exact: exact });
    }
};
exports.default = You;
module.exports = You;
